@@ -2,43 +2,81 @@ from collections import Counter
 import datetime
 import decimal
 from pathlib import Path
-from typing import Dict, List, Optional
+from typing import Dict, List, Optional, Union
 
 from jinja2 import Environment, FileSystemLoader, StrictUndefined
-from pydantic import BaseModel, validator, conlist, Extra
+from pydantic import BaseModel, validator, Extra
 from sqlalchemy import MetaData, Table
-from sqlalchemy.engine import reflection
 from sqlalchemy.sql import sqltypes
+import sqlalchemy
 
 from ..core.errors import DBError, Exc, Ok
 
 from ..utils.misc import group_list
 
 
+class Hook(BaseModel):
+    sql: str
+
+
+class Columns(BaseModel):
+    name: str
+    type: Optional[str] = None
+    dst_name: Optional[str] = None
+    description: Optional[str]
+
+    class Tests(BaseModel):
+        name: Optional[str]
+        values: Optional[List[str]]
+        execute: bool = True
+
+        class Config:
+            extra = Extra.forbid
+
+    tests: Optional[List[Union[str, Tests]]] = list()
+
+    class Config:
+        extra = Extra.forbid
+
+
 class DDL(BaseModel):
-    class Column(BaseModel):
-        name: str
-        type: Optional[str]
-        primary: Optional[bool] = False
-        not_null: Optional[bool] = False
-        unique: Optional[bool] = False
-        dst_name: Optional[str]
+    class Properties(BaseModel):
+        class Indexes(BaseModel):
+            columns: Optional[List[Union[str, Dict]]]
+
+        indexes: Optional[Dict[str, Indexes]]
 
         class Config:
             extra = Extra.forbid
 
-    class Index(BaseModel):
-        columns: conlist(str, min_items=1)
+        @validator("indexes")
+        def index_columns_exists(cls, v, values):
+            cols = [c.name for c in values.get("columns", list())]
+            if len(cols) > 0:
+                missing_cols = group_list(
+                    [
+                        (index_name, index_column)
+                        for index_name, index in v.items()
+                        for index_column in index.columns
+                        if index_column not in cols
+                    ]
+                )
+                if len(missing_cols) > 0:
+                    cols_msg = ";".join(
+                        [f"On {i}: {','.join(c)}" for i, c in missing_cols.items()]
+                    )
+                    raise ValueError(
+                        f"Some indexes refer to missing columns: {cols_msg}"
+                    )
 
-        class Config:
-            extra = Extra.forbid
+            return v
 
-    columns: Optional[List[Column]] = list()
-    indexes: Optional[Dict[str, Index]] = dict()
-    # logic field - i.e. not added by the user in the ddl definition
-    primary_key: Optional[List[str]] = []
+    columns: Optional[List[Columns]] = list()
+    properties: Optional[List[Properties]] = list()
+    post_hook: Optional[List[Hook]] = list()
 
-    permissions: Optional[Dict[str, str]] = dict()
+    class Config:
+        extra = Extra.forbid
 
     @validator("columns", pre=True)
     def transform_str_cols(cls, v, values):
@@ -55,56 +93,144 @@ class DDL(BaseModel):
         else:
             return v
 
-    @validator("indexes")
-    def index_columns_exists(cls, v, values):
-        cols = [c.name for c in values.get("columns", list())]
-        if len(cols) > 0:
-            missing_cols = group_list(
-                [
-                    (index_name, index_column)
-                    for index_name, index in v.items()
-                    for index_column in index.columns
-                    if index_column not in cols
-                ]
-            )
-            if len(missing_cols) > 0:
-                cols_msg = ";".join(
-                    [f"On {i}: {','.join(c)}" for i, c in missing_cols.items()]
-                )
-                raise ValueError(f"Some indexes refer to missing columns: {cols_msg}")
-
-        return v
-
-    @validator("primary_key", always=True)
-    def set_pk(cls, v, values):
-        columns_pk = [c.name for c in values.get("columns", []) if c.primary]
-
-        indexes_pk = list()
-        if values.get("indexes", {}).get("primary_key") is not None:
-            indexes_pk = values.get("indexes").get("primary_key").columns
-
-        if len(columns_pk) > 0 and len(indexes_pk) > 0:
-            if set(columns_pk) != set(indexes_pk):
-                columns_pk_str = " ,".join(columns_pk)
-                indexes_pk_str = " ,".join(indexes_pk)
-                raise ValueError(
-                    f"Primary key defined in indexes ({indexes_pk_str}) does not match primary key defined in columns ({columns_pk_str})."
-                )
-
-        pk = columns_pk if len(columns_pk) > 0 else indexes_pk
-
-        return pk
-
     def get_ddl(self):
-        return {
-            "columns": [c.dict() for c in self.columns],
-            "indexes": {
-                k: v.dict() for k, v in self.indexes.items() if k != "primary_key"
-            },
-            "permissions": self.permissions,
-            "primary_key": self.primary_key,
+        cols = self.columns
+        self.columns = []
+        for c in cols:
+            tests = []
+            for t in c.tests:
+                if isinstance(t, str):
+                    tests.append({"type": t, "values": [], "execute": True})
+                else:
+                    tests.append(
+                        {
+                            "type": t.name if t.name is not None else "values",
+                            "values": t.values if t.values is not None else [],
+                            "execute": t.execute,
+                        }
+                    )
+            self.columns.append(
+                {
+                    "name": c.name,
+                    "description": c.description,
+                    "dst_name": c.dst_name,
+                    "type": c.type,
+                    "tests": tests,
+                }
+            )
+
+        props = self.properties
+        self.properties = []
+        for p in props:
+            self.properties.append(p.dict())
+
+        hook = self.post_hook
+        self.post_hook = []
+        for h in hook:
+            self.post_hook.append(h.dict())
+
+        res = {
+            "columns": self.columns,
+            "properties": self.properties,
+            "post_hook": self.post_hook,
         }
 
+        return res
+
+
+class DbObject:
+    def __init__(
+        self, connection_name, database, schema, object, stringify, prod_stringify
+    ):
+        self.connection_name = connection_name
+        self.database = database
+        self.schema = schema
+        self.object = object
+        self.stringify = stringify
+        self.prod_stringify = prod_stringify
+
+        self.database_value = None
+        self.database_prod_value = None
+        self.schema_value = None
+        self.schema_prod_value = None
+        self.object_value = None
+        self.object_prod_value = None
+        self.value = ""
+        self.prod_value = ""
+
+        if self.database is not None:
+            self.database_value = self.stringify["database"].compile(
+                database=self.database, connection=connection_name
+            )
+            self.value += self.database_value + "."
+
+            self.database_prod_value = self.prod_stringify["database"].compile_prod(
+                database=self.database, connection=connection_name
+            )
+            self.prod_value += self.database_prod_value + "."
+
+        if self.schema is not None:
+            self.schema_value = self.stringify["schema"].compile(
+                table=self.object, schema=self.schema, connection=connection_name
+            )
+            self.value += self.schema_value + "."
+
+            self.schema_prod_value = self.prod_stringify["schema"].compile_prod(
+                table=self.object, schema=self.schema, connection=connection_name
+            )
+            self.prod_value += self.schema_prod_value + "."
+
+        self.object_value = self.stringify["table"].compile(
+            table=self.object, schema=self.schema, connection=connection_name
+        )
+        self.value += self.object_value
+
+        self.object_prod_value = self.prod_stringify["table"].compile_prod(
+            table=self.object, schema=self.schema, connection=connection_name
+        )
+        self.prod_value += self.object_prod_value
+
+    def __hash__(self):
+        return hash(self.get_key())
+
+    def __eq__(self, obj):
+        return self.get_key() == obj.get_key()
+
+    def __lt__(self, obj):
+        return self.get_key() > obj.get_key()
+
+    def __repr__(self):
+        return self.get_key()
+
+    def get_key(self):
+        return f"{self.connection_name}:{self.database or ''}.{self.schema or ''}.{self.object}"
+
+    def get_value(self):
+        return self.value
+
+    def get_prod_value(self):
+        return self.prod_value
+
+
+class DBObjectFactory:
+    def __init__(self, connection_name, stringify, prod_stringify):
+        self.connection_name = connection_name
+        self.stringify = stringify
+        self.prod_stringify = prod_stringify
+
+    def from_components(self, database, schema, object):
+        return DbObject(
+            self.connection_name,
+            database,
+            schema,
+            object,
+            self.stringify,
+            self.prod_stringify,
+        )
+
+    # def from_string(self, obj):
+    #     return DbObject(connection_name, database, schema, obj, self.stringify, self.prod_stringify)
+
 
 class Database:
     """
@@ -121,21 +247,32 @@ class Database:
         metadata (sqlalchemy.MetaData): A metadata object associated with the engine.
     """
 
-    ddl_validation_class = DDL
+    DDL = DDL
+    DBObjectFactory = DBObjectFactory
 
-    _requested_objects = None
-
-    def __init__(self, name, name_in_settings, db_type, common_params):
+    def __init__(
+        self, name, name_in_settings, db_type, common_params, stringify, prod_stringify
+    ):
         self.name = name
         self.name_in_settings = name_in_settings
         self.db_type = db_type
         self.max_batch_rows = common_params.get("max_batch_rows", 50000)
         self._requested_objects = dict()
+
         self._jinja_env = Environment(
             loader=FileSystemLoader(Path(__file__).parent / "templates"),
             undefined=StrictUndefined,
             keep_trailing_newline=True,
         )
+        self._jinja_test = Environment(
+            loader=FileSystemLoader(Path(__file__).parent.parent / "tasks/tests"),
+            undefined=StrictUndefined,
+            keep_trailing_newline=True,
+        )
+
+        self._object_builder = self.DBObjectFactory(
+            self.name, stringify, prod_stringify
+        )
 
     def feature(self, feature):
         # Supported sql_features
@@ -159,18 +296,102 @@ class Database:
         self.engine = engine
         self.metadata = MetaData(self.engine)
 
-        # Force a query to test the connection
-        engine.execute("select 1")
+    def _construct_tests(self, columns, table, schema=None):
+        query = """
+                   SELECT val
+                        , col
+                        , cnt
+                        , type
+                     FROM (
+                """
+        template = self._jinja_test.get_template("standard_tests.sql")
+        count_tests = 0
+        breakdown = []
 
-    def _validate_ddl(self, ddl):
-        if ddl is None or len(ddl) == 0:
-            return Ok(self.ddl_validation_class().get_ddl())
+        for col in columns:
+            tests = col["tests"]
+            for t in tests:
+                breakdown.append(
+                    {
+                        "column": col["name"],
+                        "type": t["type"],
+                        "status": col[t["type"]],
+                        "execute": t["execute"],
+                    }
+                )
+
+                if col[t["type"]] is False and t["execute"]:
+                    count_tests += 1
+                    query += template.render(
+                        **{
+                            "table": table,
+                            "schema": schema,
+                            "name": col["name"]
+                            if not col["dst_name"]
+                            else col["dst_name"],
+                            "type": t["type"],
+                            "values": ", ".join(f"'{c}'" for c in t["values"]),
+                        },
+                    )
+            breakdown.append({"print": True})
+        breakdown = breakdown[:-1]
+        parts = query.splitlines()[:-2]
+        query = ""
+        for q in parts:
+            query += q.strip() + "\n"
+        query += ") AS t;"
+
+        if count_tests == 0:
+            return Ok(["", breakdown])
+
+        return Ok([query, breakdown])
+
+    def _validate_ddl(self, columns=[], table_properties=[], post_hook=[]):
+        if len(columns) == 0 and len(table_properties) == 0 and len(post_hook) == 0:
+            return self._format_properties(self.DDL().get_ddl())
         else:
             try:
-                return Ok(self.ddl_validation_class(**ddl).get_ddl())
+                return self._format_properties(
+                    self.DDL(
+                        columns=columns,
+                        properties=table_properties,
+                        post_hook=post_hook,
+                    ).get_ddl()
+                )
             except Exception as e:
+
                 return Exc(e, db=self.name, type=self.db_type)
 
+    def _format_properties(self, properties):
+
+        if properties["columns"]:
+            columns = []
+            for col in properties["columns"]:
+                entry = {
+                    "name": col["name"],
+                    "type": col["type"],
+                    "dst_name": col["dst_name"],
+                    "unique": False,
+                    "not_null": False,
+                    "values": False,
+                }
+                if "tests" in col:
+                    entry.update({"tests": col["tests"]})
+                    for t in col["tests"]:
+                        if t["type"] != "values" and col["type"]:
+                            entry.update({t["type"]: True})
+                columns.append(entry)
+
+            properties["columns"] = columns
+
+        if properties["properties"]:
+            for pro in properties["properties"]:
+                for p in pro:
+                    if pro[p] is not None:
+                        properties[p] = pro[p]
+
+        return Ok(properties)
+
     def _refresh_metadata(self, only=None, schema=None):
         """Refreshes the sqlalchemy metadata object.
 
@@ -180,36 +401,38 @@ class Database:
         """
         self.metadata.reflect(only=only, schema=schema, extend_existing=True)
 
-    def _request_object(
-        self, name, schema=None, tmp_schema=None, task_name=None, request_tmp=True
-    ):
+    def _introspect(self, to_introspect):
+        insp = sqlalchemy.inspect(self.engine)
+        out = dict()
 
-        to_request = [(name, schema)]
-        if request_tmp:
-            to_request.append((tmp_name(name), tmp_schema or schema))
-
-        for name, schema in to_request:
-            if schema not in self._requested_objects:
-                self._requested_objects[schema] = {name: {"tasks": list()}}
+        for schema, req_objs in to_introspect.items():
+            if schema == "":
+                schema = None
+            if schema is None:
+                db_objects = [
+                    ("table", insp.get_table_names()),
+                    ("view", insp.get_view_names()),
+                ]
             else:
-                self._requested_objects[schema][name] = {"tasks": list()}
+                db_objects = [
+                    ("table", insp.get_table_names(schema)),
+                    ("view", insp.get_view_names(schema)),
+                ]
 
-            if task_name is not None:
-                self._requested_objects[schema][name]["tasks"].append(task_name)
+            # flatten the results
+            db_objects = {o: t for t, obs in db_objects for o in obs}
 
-    def _introspect(self):
-        insp = reflection.Inspector.from_engine(self.engine)
+            if schema not in self._requested_objects:
+                self._requested_objects[schema] = dict()
 
-        for schema in self._requested_objects.keys():
-            for obj_type, objs in [
-                ("table", insp.get_table_names(schema=schema)),
-                ("view", insp.get_view_names(schema=schema)),
-            ]:
-                for obj_name in objs:
-                    if schema is not None and obj_name.startswith(schema + "."):
-                        obj_name = obj_name[len(schema + ".") :]
-                    if obj_name in self._requested_objects[schema]:
-                        self._requested_objects[schema][obj_name]["type"] = obj_type
+            out[schema] = dict()
+            for obj_name in req_objs:
+                if obj_name in db_objects:
+                    out[schema][obj_name] = {"type": db_objects[obj_name]}
+                else:
+                    out[schema][obj_name] = {"type": None}
+
+        self._requested_objects = out
 
     def _py2sqa(self, from_type):
         python_types = {
@@ -332,6 +555,7 @@ class Database:
         buffer = list()
 
         result = self._validate_ddl(ddl)
+
         if result.is_err:
             raise DBError(
                 self.name,
@@ -385,6 +609,7 @@ class Database:
         Returns:
             sqlalchemy.Table: A table object from sqlalchemy
         """
+
         table_def = Table(table, self.metadata, schema=schema, extend_existing=True)
 
         if table_def.exists():
@@ -499,6 +724,7 @@ class Database:
         tmp_schema=None,
         **ddl,
     ):
+
         # Create the temporary table
         can_replace_table = self.feature("CAN REPLACE TABLE")
 
@@ -534,6 +760,8 @@ class Database:
         table_exists = object_type == "table"
         view_exists = object_type == "view"
 
+        # ddl = self._format_properties(ddl).value
+
         template = self._jinja_env.get_template("create_view.sql")
         create = template.render(
             table_name=view_name,
@@ -567,6 +795,9 @@ class Database:
 
         return {"Create Table": create_or_replace, "Merge Tables": merge}
 
+    def get_db_object(self, database, schema, object):
+        return self._object_builder.from_components(database, schema, object)
+
 
 def fully_qualify(name, schema=None):
     return f"{schema+'.' if schema is not None else ''}{name}"
