@@ -23,9 +23,8 @@ CREATE TABLE {{ full_name }}
 (
 {%- for col_def in columns %}
     {{ col_def['name'] }} {{ col_def['type'] }}
-    {{- ' PRIMARY KEY' if col_def.get('primary')}}
     {{- ' UNIQUE' if col_def.get('unique')}}
-    {{- ' NOT NULL' if col_def.get('unique')}}
+    {{- ' NOT NULL' if col_def.get('not_null')}}
     {{- ',' if not loop.last else ''}}
 {%- endfor %}
 )
@@ -41,12 +40,12 @@ CLUSTER BY {{ cluster|join(', ') }}
 {% endif %}
 
 {%- if distribution is defined and distribution is not none %}
-DISTSTYLE {{ distribution['style'] }}
-{% if distribution['style'] == 'key' %}DISTKEY({{ distribution['key'] }}){% endif %}
+DISTSTYLE {{ distribution['type'] }}
+{% if distribution['type'] == 'KEY' %}DISTKEY({{ distribution['column'] }}){% endif %}
 {% endif %}
 
 {%- if sorting is defined and sorting is not none %}
-{{ sorting['type']+' ' if 'type' in sorting else '' }}SORTKEY({{ sorting['columns']|join(', ') }})
+{{ sorting['type']+' ' if sorting['type']  else '' }}SORTKEY({{ sorting['columns']|join(', ') }})
 {% endif %}
 {% endblock -%}
 
@@ -58,7 +57,7 @@ INSERT INTO {{ full_name }}
 {% else %}
 AS
 {% endif %}
-  {%- if columns is defined and columns|length > 0 %}
+  {%- if columns is defined and columns|length > 0 and all_columns_have_type %}
 SELECT {{ columns|join('\n     , ', attribute='name') }}
   FROM ({{ select }}) t
   {%- else %}
